// A library of routines for supporting serial communications
// on the PIC 16F877A and similiar processors using the hardware UART
// used on severl different projects as well as SerialDemo.c which
// demonstrates its features.  Anything not tested there should be
// viewed with more than mormal suspission
// note that short and int seem to be interchangable if a routine is written for a short it works fine
// and will be linked in for an int as well, this make sense since both are 2 byte values
// test with DemoSerial.c
//
// History:
//		Last author russ_hensel http://www.opencircuits.com/User:Russ_hensel
//		russ_hensel version, source location ...DemoSerial\Ver3
//		Prior history 
//			Lieven Hollevoet
//				most recent version on http://boostc.lika.be
//			Based on code from Yann Hendrikx.
// 
// Environment:  
//		SourceBoost BoostC  6.81 full version on XP
//		PIC: 16f877A



#include <system.h>
#include <string.h>
#include <stdio.h> 

#include "SerialLib.h"

// =========== subroutine ==========
// Hardware initialisation
// input argument: the baudrate scaler

void serial_init877( char brg )
{
	// rh revision for 877 which may or may not differ from 18f4620
	// comments only as code did not change.
	// Enable USART function on PORTC (see PIC18F4620 datasheet page 201)
	// Bit SPEN (RCSTA<7>) and bits TRISC<7:6> have to
	// be set in order to configure pins RC6/TX/CK and
	// RC7/RX/DT as the Universal Synchronous Asynchronous Receiver Transmitter.
	
	trisc.7       = 1;			
	trisc.6       = 1; 		
	
	// Transmit control register  TXSTA:
	txsta.TX9     = 0;    	//	TX9: 9-bit Transmit Enable bit
							//	1 = Selects 9-bit transmission
							//	0 = Selects 8-bit transmission
	txsta.TXEN    = 1;		//	bit 5  TXEN: Transmit Enable bit
							//	1 = Transmit enabled
							//	0 = Transmit disabled
	txsta.SYNC    = 0; 		//	SYNC: USART Mode Select bit
							//	1 = Synchronous mode
							//	0 = Asynchronous mode
	txsta.BRGH    = 1;		//	bit 2 BRGH: High Baud Rate Select bit
							//	Asynchronous mode:
							//	1 = High speed
							//	0 = Low speed
	
	// Receive control register
	rcsta.SPEN    = 1;		// 	Bit SPEN (RCSTA<7>) and bits TRISC<7:6> have to
							//	be set in order to configure pins RC6/TX/CK and
	rcsta.RX9     = 0;		// 	RX9: 9-bit Receive Enable bit
							//	1 = Selects 9-bit reception
							//	0 = Selects 8-bit reception
	rcsta.CREN    = 1;		//	CREN: Continuous Receive Enable bit
							//	Asynchronous mode:
							//	1 = Enables continuous receive
							// 	0 = Disables continuous receive
	rcsta.ADDEN   = 0;		//	ADDEN: Address Detect Enable bit
							//	Asynchronous mode 9-bit (RX9 = 1):
							//	1 = Enables address detection, enables interrupt and load of the receive buffer when 
							//	RSR<8> is set
							//	0 = Disables address detection, all bytes are received, and ninth bit can be used as parity bit
	
	// Baudcon register, if supported by the target architecture
#ifdef baudcon
	// baudcon not found in the manual
	baudcon.BRG16 = 0;
	baudcon.WUE   = 0;
	baudcon.ABDEN = 0;
#endif
	// Baud rate setting
    spbrg 	   = brg;	
 }
 
// =========== subroutine ==========
 void serial_init(char brg)
{
	// Enable USART function on PORTC (see PIC18F4620 datasheet page 201)

	trisc.7       = 1;			
	trisc.6       = 1; 	
		
	//txsta	= 00100100b;        //full duplex asynchronous a one line version 
	
	// Transmit control register
	txsta.TX9     = 0;
	txsta.TXEN    = 1;
	txsta.SYNC    = 0;
	txsta.BRGH    = 1;
	
	// Receive control register
	rcsta.SPEN    = 1;
	rcsta.RX9     = 0;
	rcsta.CREN    = 1;
	rcsta.ADDEN   = 0;
	
	// Baudcon register, if supported by the target architecture
#ifdef baudcon
	baudcon.BRG16 = 0;
	baudcon.WUE   = 0;
	baudcon.ABDEN = 0;
#endif
	// Baud rate setting
    spbrg 	   = brg;	
 }


// =========== subroutine ==========
// Print a single character to the serial port
// This funtion is skipped in debug mode since it will 
// block the debugger (the function waits for a hardware
// signal)
// Arguments:
//		value  -> the character to be sent
//		inteacts with uart registers

void serial_printf( char value )   {

#ifdef DBG_COMPILE
	// Skip rest of the routine when debugging
	#warning "Debug build, serial_printf will return immediately"
	return;
#endif

    while( ( txsta & 1 << TRMT ) == 0 ) {
		// null -- loop waiting for transmitter to be ready
    }
    txreg = value;
    return;
}

// =========== subroutine ==========
// Prints a line to the serial port. 
// The function parameter is stored in rom.
// russ:  think may be in rom or not
// Arguments:
//		text -> the string to print

void serial_printf( const char* text )  {
    char i = 0;
    while(text[i] != 0)  {
        serial_printf( text[i++] );
	}
	// return;
}

// =========== subroutine ==========
// Print as a decimal value to the serial port
// note that there are several versions of this function for different call types
// Arguments:
//		number -> the number to print

void serial_print_dec( unsigned char number)  {
    if(number > 99) {
        serial_printf(((number / 100) % 10) + '0');
    }
    if(number > 9) {
        serial_printf(((number / 10) % 10) + '0');
    }
    serial_printf((number % 10) + '0');
    // return;  -- build considers this rundundant
}

// =========== subroutine ==========
// Print as a decimal value to the serial port
// note that there are several versions of this function for different call types
// Arguments:
//		number -> the number to print

void serial_print_dec( signed char number )  {
	
	char  usnumber;

	if ( number < 0 ) { 
		serial_printf( "-" );
		usnumber = -number;
	} else {
		usnumber = number;
	}
		
    serial_print_dec( usnumber );
    
    // return;  -- build considers this rundundant
}

// =========== subroutine ==========
// Print a decimal value to the serial port
// The number takes multiple characters. If a character 
// position is not used, it is filled with 0.
// Positions should be 1, 2 or 3.

void serial_print_dec( char number, char positions )  {

    if(number > 99) {
        serial_printf(((number / 100) % 10) + '0');
    } else {
		if (positions > 2){
			serial_printf('0');
		}
	}
    if(number > 9) {
        serial_printf(((number / 10) % 10) + '0');
    } else {
		if (positions > 1){
			serial_printf('0');
		}
	}
	serial_printf((number % 10) + '0');
	// return;  -- build considers this rundundant
}

// =========== subroutine ==========
// Print as a decimal value to the serial port
// note that there are several versions of this function for different call types
// works to 65000 or 65k ( actually FFFF )
// Arguments:
//		number -> the number to print

void serial_print_dec( unsigned short number ) {

    
	if(number > 9999) {
        serial_printf(((number / 10000) % 10) + '0');
    }
	if(number > 999) {
        serial_printf(((number / 1000) % 10) + '0');
    }
    if(number > 99) {
        serial_printf(((number / 100) % 10) + '0');
    }
    if(number > 9) {
        serial_printf(((number / 10) % 10) + '0');
    }
    serial_printf((number % 10) + '0');
    
    // return;  -- build considers this rundundant
}

// =========== subroutine ==========
// Print as a decimal value to the serial port
// note that there are several versions of this function for different call types
// works to 65000 or 65k ( actually FFFF )
// Arguments:
//		number -> the number to print

void serial_print_dec( signed short number)  {
	
	unsigned short  usnumber;

	if ( number < 0 ) { 
		serial_printf( "-" );
		usnumber = -number;
	} else {
		usnumber = number;
	}
		
    serial_print_dec( usnumber );    
    // return;  -- build considers this rundundant
}

// ----------- end decimal routines -----------
// ----------- begin hex routines -----------

// =========== subroutine ==========
// Print as a hex  value to the serial port
// note that there are differnt versions of this function for different call types
// Arguments:
//		number -> the number to print

void serial_print_hex( unsigned char number ) {
    char hexChar;
    char i;
    for(i = 0; i < 2; i++)      {
        if(i == 0)
            hexChar = number >> 4;
        else
            hexChar = number & 0x0F;
        if(hexChar < 10)
            hexChar = hexChar + '0';
        else
            hexChar = hexChar + ('A' - 10 );
        serial_printf(hexChar);
    }
}

// =========== subroutine ==========
// Print as a hex  value to the serial port
// note that there are differnt versions of this function for different call types
// Arguments:
//		number -> the number to print

void serial_print_hex( unsigned short value ){
	
    char value1 = (char)((value >> 8) & 0x00FF);
    
	serial_print_hex( value1 );
 
    char value0 = ( char )( value & 0x00FF );
    
	serial_print_hex( value0 );
	
	// return;  -- build considers this rundundant
}

// ----------- end hex routines -----------

// =========== subroutine ==========
// Print a binary value to the serial port
// will print leading zeros, 0 prints as 00000000

void serial_print_bin( char number )
{
    char i;
    for(i = 0; i < 8; i++)
    {
		if (number & 10000000b){
			serial_printf('1');
		} else {
		    serial_printf('0');
		}
        number <<= 1;
    }
}
 
// =========== subroutine ==========
// Wait for a character to read it from the serial port
// This function is skipped when running in debug mode
// since else it blocks the simulator (waits for a 
// hardware signal from the UART)

char serial_getch(){

#ifdef DBG_COMPILE
	#warning "Debug build, serial_getch will return 0x00"
	return 0x00;
#endif
  
	// Check for overrun error. Reset the 
	// UART if any.
	if (rcsta.OERR){
		serial_printf("OERR");
		rcsta.CREN = 0;
		rcsta.CREN = 1;
	}
	
	while (!serial_peek());
	return rcreg;
}

// =========== subroutine ==========

// Read a decimal value from the serial port
// needs to be terminated with 0x0D  = cr  looks like cr lf could be a problem ! switch to lf = 0x0A
// echos back ( add as flag? )
// assumes that chars are comming in as  digits one after the other, not bufferd
// converted as they come in -- since ret is a char max is 255 or so
// not tested 

char serial_get_decimal(){
	char value = 0;
	char input = 0;

	input = serial_getch();
	
	while (input != 0x0D){ 
		//serial_printf(input);  // remove to get rid of echo
		value *= 10;
		value += (input-0x30);   // 0x30 = char 0

		input = serial_getch();
	}
	serial_print_lf();
	return value;
}


// ========= begin command building routines =========


// ----- variables for command interface 
unsigned  	char 	CmdIx;
unsigned  	char 	Cmd[CMD_MAX];   // can init with = "ss" but cannot reassign later
unsigned  	char 	CmdPrior[CMD_MAX];

unsigned	char 	stopFlag;	 //true when we have gotten a stop command

unsigned  	char  	cmdStatus;

unsigned 	char	GotCmd;


// ---------- begin subroutine ----------
// add a char to the string
// note: not really a string until it has been ended with a 0
// arguments
//		aNewChar	call: the character to be added
//		globals:	may change a bunch of the command interface variables
void   buildCmd( unsigned char aNewChar )  {

		Cmd[CmdIx] =  aNewChar;
		CmdIx ++;
		if ( CmdIx >= CMD_MAX ) {
			CmdIx = CMD_MAX;
		}
}

// ---------- begin subroutine ----------
// eliminate in favor of serial.c serial_printf()
//void sendString( const char* text )   {
//
//	char i = 0;
//	while( text[i] != 0 ) {
//		sendByte( text[i++] );
//	}
//}
 
// ---------- begin subroutine ----------
// backspace the command
// arguments
//		globals:	may change a bunch of the command interface variables
void   backCmd(  )  {

		if ( CmdIx > 0 ) {
			CmdIx --;
		}
		// return;  -- build considers this redundant
}
  
// ---------- begin subroutine ----------
// reset the command to null, inline function good?
// arguments
//		globals:	may change a bunch of the command interface variables
void resetCmd(  )  {

		CmdIx 			= 0;
		GotCmd 			= 0;
		cmdStatus 		= CMD_STATUS_ADD;
}

// ---------- begin subroutine ----------
//void sendString( const char* text )   {

//	char i = 0;
//	while( text[i] != 0 ) {
//		sendByte( text[i++] );
//	}
//}

// ---------- begin subroutine ----------
//Function: Sends a unsigned character converted ASCII decimal
// todo*   hardcode cmdprior?
int	 parseCmdNbr(  const char* text )    {

	int  Nbr = 0;
	char next;
	char ix;
	
	for( ix = 1; ix < 8; ix++ ) 	{
		next = text[ix];
		if( next == 0 ) {
			return Nbr;
		}
		Nbr = ( Nbr * 10 ) + ( next - 48 );
	} 
	return Nbr;
}



// ---------- begin subroutine ----------
// try to recieve some data 
// may use as interrupt service routine in which case serial peek test not required
void tryRecCmd ( ) {

		unsigned char data_in;
		
		if ( !serial_peek()){
			return;
		}
		
		// add check for overrun as in serial.c?
		
        data_in 	= rcreg;
//        newdata 	= 1;
        rcreg 		= 0;
        
        if ( data_in == CHAR_LF ) {
			// ignore this char
			return;
        }
        
        if ( data_in == STOP_CHAR )  {
        //if ( ( data_in == 's' ) || ( data_in == 'S' ) ) {
			// stop no matter what else we are in the process of
			// do not log as command
			stopFlag = 1;
			resetCmd(  );
			return;
        }
        
        if ( data_in == CHAR_CR ) {
			// marks end of command execute the command ( or set flag for same )
			// may want to save so can begin to build a new command
			// void strcpy( char *dst, const char *src )
			buildCmd( 0 );  // terminate the string
			strcpy( CmdPrior, Cmd );
			resetCmd(  );
			GotCmd = 1;
			return;
        }
        
        if ( data_in == CHAR_BS ) {
        //if (( data_in == CHAR_BS ) || ( data_in == CHAR_DEL )) {
			backCmd();
			return;
        }
        
        if ( cmdStatus == CMD_STATUS_ADD ) {
        
			buildCmd( data_in );
       
        }
        // if here just drop the recieved char
        // are we too long in the interrup think not
        return;
}



// =========== end ==========











